@@ -37,11 +37,13 @@ package URI::gemini {
 }
 
 package Net::Gemini;
-our $VERSION = '0.07';
+our $VERSION = '0.08';
 use strict;
 use warnings;
+use Digest::SHA 'sha256_hex';
 use Encode ();
 use Exporter 'import';
+use IO::Socket::IP;
 use IO::Socket::SSL;
 use Net::SSLeay;
 use Parse::MIME 'parse_mime_type';
@@ -53,16 +55,23 @@ sub _DEFAULT_MAX_CONTENT ()    { 2097152 }
 sub _DEFAULT_REDIRECTS ()      { 5 }
 sub _DEFAULT_REDIRECT_SLEEP () { 1 }
 
-sub code    { $_[0]{_code} }       # 0..6 response code
-sub content { $_[0]{_content} }    # NOTE only after certain calls and codes
-sub error   { $_[0]{_error} }      # error message for 0 code
-sub host    { $_[0]{_host} }
-sub meta    { $_[0]{_meta} }
-sub mime    { $_[0]{_mime} }       # NOTE only after certain calls and codes
-sub port    { $_[0]{_port} }
-sub socket  { $_[0]{_socket} }
-sub status  { $_[0]{_status} }     # two digit '1x', '2x', ... response code
-sub uri     { $_[0]{_uri} }
+sub code { $_[0]{_code} }    # 0..6 response code
+
+sub content {
+    $_[0]{_content};
+}                            # NOTE only after certain calls and codes
+sub error  { $_[0]{_error} } # error message for 0 code
+sub host   { $_[0]{_host} }
+sub ip     { $_[0]{_ip} }
+sub meta   { $_[0]{_meta} }
+sub mime   { $_[0]{_mime} }  # NOTE only after certain calls and codes
+sub port   { $_[0]{_port} }
+sub socket { $_[0]{_socket} }
+
+sub status {
+    $_[0]{_status};
+}                            # two digit '1x', '2x', ... response code
+sub uri { $_[0]{_uri} }
 
 # see VERIFICATION below; the caller should supply a custom callback.
 # the default is thus "Trust On Almost Any Use" (TOAAU) or similar to
@@ -96,7 +105,14 @@ sub get {
     # gemini://makeworld.space/gemlog/2020-07-03-tofu-rec.gmi
     # gemini://alexschroeder.ch/page/2020-07-20%20Does%20a%20Gemini%20certificate%20need%20a%20Common%20Name%20matching%20the%20domain%3F
     eval {
-        $obj{_socket} = IO::Socket::SSL->new(
+        $obj{_socket} = IO::Socket::IP->new(
+            PeerAddr => $obj{_host},
+            PeerPort => $obj{_port},
+            Proto    => 'tcp'
+        ) or die $!;
+        $obj{_ip} = $obj{_socket}->peerhost;
+        IO::Socket::SSL->start_SSL(
+            $obj{_socket},
             SSL_hostname => $obj{_host},    # SNI
             ( $param{tofu} ? ( SSL_verifycn_scheme => 'none' ) : () ),
             SSL_verify_callback => sub {
@@ -105,18 +121,25 @@ sub get {
                     return 1 if $param{tofu};
                     return $ok;
                 }
-                ( $param{verify_ssl} || \&_verify_ssl )->(
-                    @obj{qw(_host _port)},
-                    Net::SSLeay::X509_get_fingerprint( $cert, 'sha256' ),
-                    Net::SSLeay::P_ASN1_TIME_get_isotime( Net::SSLeay::X509_get_notBefore($cert) ),
-                    Net::SSLeay::P_ASN1_TIME_get_isotime( Net::SSLeay::X509_get_notAfter($cert) ),
-                    $ok,
-                    $cert
+                my $digest = ( $param{verify_ssl} || \&_verify_ssl )->(
+                    {   host   => $obj{_host},
+                        port   => $obj{_port},
+                        cert   => $cert,         # warning, memory address!
+                                                 # compatible with certID function of amfora
+                        digest =>
+                          uc( sha256_hex( Net::SSLeay::X509_get_X509_PUBKEY($cert) ) ),
+                        ip        => $obj{_ip},
+                        notBefore => Net::SSLeay::P_ASN1_TIME_get_isotime(
+                            Net::SSLeay::X509_get_notBefore($cert)
+                        ),
+                        notAfter => Net::SSLeay::P_ASN1_TIME_get_isotime(
+                            Net::SSLeay::X509_get_notAfter($cert)
+                        ),
+                        okay => $ok,
+                    }
                 );
             },
             ( exists $param{ssl} ? %{ $param{ssl} } : () ),
-            PeerHost => $obj{_host},
-            PeerPort => $obj{_port},
         ) or die $!;
         1;
     } or do {
@@ -131,17 +154,12 @@ sub get {
         @obj{qw(_code _error)} = ( 0, "send URI failed: $!" );
         goto BLESSING;
     }
-    # KLUGE we're done with the connection as a writer at this point,
-    # but IO::Socket::SSL does not appear to offer a public means to
-    # only call shutdown and nothing else. using this is a bit risky
-    # should the IO::Socket::SSL internals change
-    Net::SSLeay::shutdown( ${ *{ $obj{_socket} } }{'_SSL_object'} )
-      if $param{early_shutdown};
 
     # get the STATUS SPACE header response (and, probably, more)
     $obj{_buf} = '';
     while (1) {
-        my $n = sysread $obj{_socket}, my $buf, $param{bufsize} || _DEFAULT_BUFSIZE;
+        my $n = sysread $obj{_socket}, my $buf,
+          $param{bufsize} || _DEFAULT_BUFSIZE;
         unless ( defined $n ) {
             @obj{qw(_code _error)} = ( 0, "recv response failed: $!" );
             goto BLESSING;
@@ -161,8 +179,12 @@ sub get {
         @obj{qw(_status _code)} = ( $1, $2 );
         substr $obj{_buf}, 0, 3, '';
     } else {
-        @obj{qw(_code _error)} =
-          ( 0, "invalid response " . sprintf "%vx", substr $obj{_buf}, 0, 3 );
+        @obj{qw(_code _error)} = (
+            0,
+            "invalid response " . sprintf "%vx",
+            substr $obj{_buf},
+            0, 3
+        );
         goto BLESSING;
     }
 
@@ -180,7 +202,8 @@ sub get {
                 $obj{_meta} = 'text/gemini;charset=utf-8' if $obj{_code} == 2;
             } else {
                 eval {
-                    $obj{_meta} = Encode::decode( 'UTF-8', $obj{_meta}, Encode::FB_CROAK );
+                    $obj{_meta} =
+                      Encode::decode( 'UTF-8', $obj{_meta}, Encode::FB_CROAK );
                     1;
                 } or do {
                     @obj{qw(_code _error)} = ( 0, "failed to decode meta: $@" );
@@ -246,7 +269,8 @@ sub gemini_request {
         $gem->{_mime} = [ parse_mime_type( $gem->meta ) ];
         if ( exists $options{content_callback} ) {
             if ($len) {
-                $options{content_callback}->( $gem->{_buf}, $len, $gem ) or goto CLEANUP;
+                $options{content_callback}->( $gem->{_buf}, $len, $gem )
+                  or goto CLEANUP;
             }
             while (1) {
                 my $buf;
@@ -462,15 +486,6 @@ Parameters include:
 Size of buffer to use for requests, 4096 by default. Note that a naughty
 server may return data in far smaller increments than this.
 
-=item B<early_shutdown> => I<boolean>
-
-If true, attempts an early shutdown of the SSL connection after the
-request is sent. This fiddles with the internal state of
-L<IO::Socket::SSL> as the C<shutdown> call does not appear to be exposed
-outside the module.
-
-Use with caution.
-
 =item B<ssl> => { params }
 
 Passes the given parameters to the L<IO::Socket::SSL> constructor. These
@@ -496,27 +511,25 @@ the B<ssl> param if this needs to be customized.
 
 Custom callback function to handle SSL verification. The default is to
 accept the connection (Trust On All Uses), which is perhaps not ideal.
-The callback function is passed the host, port, certificate digest, and
-certificate expiration date (compatible with
-L<DateTime::Format::RFC3339>) and should return a C<1> to verify the
-connection, or C<0> to not.
+The callback is passed a hash reference containing various information
+about the certificate and connection.
 
   ...->get( $url, ..., verify_ssl => sub {
-    my ($host, $port, $digest, $not_before, $not_after, $ok, $raw_cert) = @_;
+    my ($param) = @_;
     return 1 if int rand 2; # certificate is OK
     return 0;
   } );
 
-The "okay?" boolean and raw certificate is also passed; these could be
-used to allow certificates that other code was able to verify, or to
-perform custom checks on the certificate using probably various routines
-from L<Net::SSLeay>.
-
 Note that some have argued that under TOFU one should not verify the
 hostname nor the dates (notBefore, notAfter) of the certificate, only to
 accept the first certificate presented as-is, like SSH does, and to use
 that certificate thereafter. This has plusses and minuses.
 
+See C<bin/gmitool> for how C<verify_ssl> might be used in a client.
+
+In module version 0.08 the format of the digest (fingerprint) changed to
+be compatible with the amfora gemini client.
+
 =back
 
 =item B<getmore> I<callback> [ bufsize => n ]
@@ -566,6 +579,10 @@ The error message, if any.
 
 Host used for the request.
 
+=item B<ip>
+
+IP address used for the request.
+
 =item B<meta>
 
 Gemini meta line. Use varies depending on the code.
